/********************************************************************
 *
 *	TITLE:	   x16io
 *
 *	DATE:	   05-26-82
 *
 *	ABSTRACT:  .
 *
 *	LANGUAGE DEPENDENCIES:
 *
 *		C86     VERSION 1.0
 *
 *********************************************************************/
#include <curses.h>
#include <errno.h>
#include "common.def"

	/*
	 *  Some definitions
	 */

#define	READ_WRITE	2
#define RW_R		0644

#define	CUR_VALUE	0
#define	COMPARE		1
#define	CLOSE		2
#define	CREATE		3
#define	UNLINK		4
#define	OPEN		5
#define	READ		6
#define	SEEK		7
#define	WRITE		8

#define	SPEED_TRAP	20
#define	PTRS		21
#define	ALL		22

struct test_desc {
	struct msg_dsc msgs;
	char path[NAMEMAX];
	char *seg;
	char value;
	char odometer;
	bool fill;
	int status;
	int fd;
	unsigned size;
	int pass;
	int line;
	long file_ptr;
};

	/*
	 *  External variables
	 */

extern int errno;
extern int go[];
extern unsigned buf_size;
extern int speed;
extern int synch[];
extern int t_line;
extern char path[];

/********************************************************************
 *
 *	TITLE:	   display
 *
 *	CALLING SEQUENCE:
 *
 *		   display( test, d_code );
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
display( test, d_code )
struct test_desc *test;
unsigned short d_code;
{


	if ( d_code <= WRITE ) {
		if ( speed <= CRUISE ) {
			switch( d_code ) {
				case 0:	sprintf( test->msgs.msg, "%4x", (short) test->value & 0xff );
					xadd( &test->msgs, test->line, 58 );
					break;
				case 1:	xaddstr( &test->msgs, test->line, 66, "compare......" );
					break;
				case 2:	xaddstr( &test->msgs, test->line, 66, "close........" );
					break;
				case 3:	xaddstr( &test->msgs, test->line, 66, "create file.." );
					break;
				case 4:	xaddstr( &test->msgs, test->line, 66, "unlink file.." );
					break;
				case 5:	xaddstr( &test->msgs, test->line, 66, "open........." );
					break;
				case 6:	xaddstr( &test->msgs, test->line, 66, "read........." );
					break;
				case 7:	xaddstr( &test->msgs, test->line, 66, "seek........." );
					break;
				case 8:	xaddstr( &test->msgs, test->line, 66, "write........" );
					break;
			}
			xrefresh( &test->msgs );
		}
		return;
	}

	if ( d_code == SPEED_TRAP ) {
		switch( speed ) {
			case CRUISE:	d_code = PTRS;
					break;
			case D_NICKLE:	if ( ++test->odometer % 55 == 0 )
						d_code = PTRS;
					break;
			case OH_WOW:	return;
		}
	}

	if ( d_code >= PTRS ) {
		sprintf( test->msgs.msg, "%9ld", test->file_ptr );
		xadd( &test->msgs, test->line, 46 );

		sprintf( test->msgs.msg, "%4x", (short) test->value & 0xff );
		xadd( &test->msgs, test->line, 58 );

		if ( d_code == ALL ) {
			sprintf( test->msgs.msg, "%5d", test->pass );
			xadd( &test->msgs, test->line, 30 );

			if ( test->fill )
				xaddstr( &test->msgs, test->line, 38, "Fill " );
			else
				xaddstr( &test->msgs, test->line, 38, "Check" );

			/*
			 *  This should go with the display of the
			 *  File name and buffer size in the
			 *  initialization section.  But, if variable
			 *  speed is implemented it will end up here.
			 */

			if ( speed == D_NICKLE )
				xaddstr( &test->msgs, test->line, 66, "Double nickle" );
			else {
				if ( speed == OH_WOW )
					xaddstr( &test->msgs, test->line, 66, "Full speed..." );
			}
		}
		xrefresh( &test->msgs );
	}

}

/********************************************************************
 *
 *	TITLE:	   check_err
 *
 *	CALLING SEQUENCE:
 *
 *		   check_err( test, code );
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
check_err( test, code )
struct test_desc *test;
int code;
{
int err_num;

	if ( test->status != E_IO ) 
		return;

	err_num = errno;
	
	display( test, ALL );

	xfstr( &test->msgs, test->line, 00, "*" );

	switch( code ) {
		case 0: xfstr(  &test->msgs, test->line, 66, "Help I'm lost " );
			break;
		case 1: xfstr(  &test->msgs, test->line, 66, "mismatch error" );
			break;
		case 2:	xfstr(  &test->msgs, test->line, 66, "close err     " );
			break;
		case 3:	xfstr(  &test->msgs, test->line, 66, "create err    " );
			break;
		case 4:	xfstr(  &test->msgs, test->line, 66, "unlink err    " );
			break;
		case 5:	xfstr(  &test->msgs, test->line, 66, "open err      " );
			break;
		case 6:	xfstr(  &test->msgs, test->line, 66, "read err      " );
			break;
		case 7:	xfstr(  &test->msgs, test->line, 66, "seek err      " );
			break;
		case 8:	xfstr(  &test->msgs, test->line, 66, "write err     " );
	}
	if ( code > 1 )  {
		sprintf( test->msgs.msg, "%2d", err_num );
		xfadd( &test->msgs, test->line, 77 );
	}

	xrefresh( &test->msgs );
	
	exit( 0 );
}

/********************************************************************
 *
 *	TITLE:	   mismatch
 *
 *	CALLING SEQUENCE:
 *
 *		   mismatch( test );
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
mismatch( test )
struct test_desc *test;
{

	display( test, COMPARE );

	if ( compare( test->value, test->seg, test->size ) != 0 ) {
		test->status = E_IO;
		check_err( test, COMPARE );
	}
}

/********************************************************************
 *
 *	TITLE:	   io_test
 *
 *	CALLING SEQUENCE:
 *
 *		   if ( fork() == 0 )
 *			io_test();
 *
 *	INTERFACE VARIABLES:
 *
 *		   .
 *              
 *	CALLS:	   .
 *
 *	ABSTRACT:  .
 *
 *
 *********************************************************************/
io_test()
{
bool eof;
char c;
struct test_desc test;


	/*
	 *  Initialize variables
	 */ 

	strcpy( test.path, path );

	test.line = t_line;

	test.size = buf_size;

	test.odometer = 0;

	test.pass = 0;

	test.file_ptr = 0;
	
	/*
	 *  Create the segment
	 */
	 
	test.seg = (char *) malloc( test.size );
	if (test.seg == NULL)
		err_msg( "Out of memory", "", FALSE );

	/*
	 *  Synch up with other io tests
	 */

	test.status = write( synch[1], "done", 4 );
	if ( test.status == E_IO )
		err_msg( "Bad child write of synch pipe", "", TRUE );


	test.status = read( go[0], &c, 1 );
	if ( test.status == E_IO )
		err_msg( "Bad child read of go pipe", "", TRUE );

	/*
	 *  Display path name and buffer size
	 */

	xaddstr( &test.msgs, test.line, 01, test.path );

	sprintf( test.msgs.msg, "%7u", test.size );
	xadd( &test.msgs, test.line, 21 );

	xrefresh( &test.msgs );

	/*
	 *  Perform the test
	 */

	for ( ; ; ) {

		eof = FALSE;
		++test.pass;
		test.fill = TRUE;
		test.value = 0;

		display( &test, ALL );

		/*
		 *  Unlink the file
		 */

		display( &test, UNLINK );
		test.status = unlink( test.path );
		if (errno != ENOENT)
			check_err( &test, UNLINK );

		/*
		 *  Create the file
		 */

		display( &test, CREATE );
		test.status = creat( test.path, RW_R );
		check_err( &test, CREATE );
		test.fd = test.status;

		/*
		 *  Close the file
		 */

		display( &test, CLOSE );
		test.status = close(test.fd);
		check_err( &test, CLOSE );

		/*
		 *  Open the file
		 */
	 
		display( &test, OPEN );
		test.status = open( test.path, READ_WRITE );
	 	check_err( &test, OPEN );
		test.fd = test.status; 

		while ( test.file_ptr <= 100000 && !eof ) {
/*
 *	 	while ( !eof ) {
 */
			++test.value;
			display( &test, CUR_VALUE );
			setch( test.value, test.seg, test.size );

			/*
			 *  Write buffer to device
			 */
	   
			display( &test, WRITE );
			test.status = write( test.fd, test.seg, test.size );
			if ( test.status == E_IO && errno == ENOSPC ) 
				eof = TRUE;
			else {
				
				/*
				 *  Check that Write was OK
				 */
				 
				 check_err( &test, WRITE );

				/*
				 *  Seek back into file
				 */
			  
				display( &test, SEEK );
				test.status = lseek( test.fd, -((long) test.size), CUR_POS );
			  	check_err( &test, SEEK ); 

				/*
				 * set buffer to complement of test.value
				 */

				setch( ~test.value, test.seg, test.size );

				/*
				 *  read data that was just written
				 */
			  
				display( &test, READ );
				test.status = read( test.fd, test.seg, test.size );
			 	check_err( &test, READ ); 

				/*
				 *  inspect contents of buffer
				 */

				mismatch( &test );

				/*
				 *  Compute new file pointer
				 */

				test.file_ptr += test.size;

				display( &test, SPEED_TRAP );

			}

	      
		}

		/*
		 *  The file is full, so let's check the data
		 */

		test.fill = FALSE;

		display( &test, ALL );

		/*
		 *  Seek to end of file
		 */

		display( &test, SEEK );
		test.status = lseek( test.fd, test.file_ptr, BEGINING );
		check_err( &test, SEEK );

		/*
		 *  Check file
		 */

		for ( ; test.file_ptr != 0; test.file_ptr -= test.size ) {
	   
			display( &test, SPEED_TRAP );

			--test.value;
			display( &test, CUR_VALUE );

			/*
			 *  Seek back into file
			 */
			  
			display( &test, SEEK );
			test.status = lseek( test.fd, -((long) test.size), CUR_POS );
		  	check_err( &test, SEEK );

			/*
			 *  Set buffer to complement of value
			 */

			setch( ~test.value, test.seg, test.size );

			/*
			 *  read data from file
			 */

			display( &test, READ );
			test.status =  read( test.fd, test.seg, test.size );
	  		check_err( &test, READ ); 

			/*
			 *  inspect contents of buffer
			 */

			mismatch( &test );

			/*
			 *  Seek back into file
			 */
			  
			display( &test, SEEK );
			test.status = lseek( test.fd, -((long) test.size), CUR_POS );
		  	check_err( &test, SEEK );

		}

		/*
		 *  Close the file
		 */

		display( &test, CLOSE );
		test.status = close( test.fd );
		check_err( &test, CLOSE );
		
	} /* forever */

}

